----------------------------------------------------------------------------
-- @Author: ViperGTS96------------------------------------------------------
----------------------------------------------------------------------------
--------------------"The simplest design is the best design." --------------
----------------------------------------------------------------------------
----------------------------------------------------------------------------

mudFX = {};
mudFX.dir = "";
mudFX.useWetness = true;
source(g_currentModDirectory.."mudFX_overwrites.lua");
local modDescFile = loadXMLFile("modDesc", g_currentModDirectory.. "modDesc.xml");
mudFX.title = getXMLString(modDescFile, "modDesc.title.en");
mudFX.author = getXMLString(modDescFile, "modDesc.author");
mudFX.version = getXMLString(modDescFile, "modDesc.version");
mudFX.printVehiclesToLog = false; --(debugging)
mudFX.overrideRain = false; --(debugging)
mudFX.useTerrainColors = false;
mudFX.maxWheelLimit = 10;
mudFX.psSpeedRatio = 5000; --Divedend
mudFX.psEmitRate = 0.4301; --Base value
mudFX.mudDensity = 0.151; --(115mud/76dirt)*0.1;
mudFX.uvSpeed = 0.01;
delete(modDescFile);
local optionsFile = loadXMLFile("options", g_currentModDirectory.. "options.xml");
mudFX.overrideRainModel = Utils.getNoNil(getXMLBool(optionsFile, "mudFX_Options.rain.overrideRainModel"),false);
mudFX.maxMudNodes = Utils.getNoNil(getXMLInt(optionsFile, "mudFX_Options.mudDirtFactor.maxMudNodes"),1800);
mudFX.printNodeCount = Utils.getNoNil(getXMLBool(optionsFile, "mudFX_Options.mudDirtFactor.printNodeCount"),false);
delete(optionsFile);

function mudFX:setDir(dir) mudFX.dir = dir; end;

function mudFX:disableWetnessCheck(disableWetness) mudFX.useWetness = not disableWetness; end;

function mudFX.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(Washable, specializations)
		and SpecializationUtil.hasSpecialization(AnimatedVehicle, specializations);
end;

function mudFX.registerEventListeners(vehicleType)
	for _, functionName in pairs( { "onPostLoad", "onUpdate", "onLeave" } ) do
		SpecializationUtil.registerEventListener(vehicleType, functionName, mudFX);
	end;
end;

function mudFX:onPostLoad(savegame)

	if mudFX.dir == "" then return; end;
	if self.spec_wheels == nil then return; end;
	local wheels = self.spec_wheels.wheels;
	local crawlers = self.spec_crawlers;
	if wheels ~= nil and #wheels > 0 then
		self.mud_FX = {};
		--self.mud_FX.initialized = false;
		self.mud_FX.tooManyWheelsForMud = false;
		local shortCrawlers = false;
		if crawlers ~= nil and #crawlers.crawlers > 0 then
			crawlers = crawlers.crawlers;
			self.mud_FX.crawlers = #crawlers;
			if #crawlers > 2 or self.spec_combine ~= nil then
				shortCrawlers = true;
			end;
		end;
		local nodeCount = 0;
		for _, washableNodes in pairs(self.spec_washable.washableNodes) do
			for _, node in pairs(washableNodes.nodes) do
				nodeCount = nodeCount+1;
			end;
		end;
		if mudFX.printNodeCount then print(mudFX.title.." : vehicle's washable node count : [ "..nodeCount.." ]"); end;
		if nodeCount <= mudFX.maxMudNodes then
			local disableMudOnVehicle = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.disableMudFXDirtFactor"),false);
			if disableMudOnVehicle then
				print(mudFX.title.." : vehicle's mud [dirt] factor disabled");
			else
				self.getDirtMultiplier = Utils.overwrittenFunction(Washable.getDirtMultiplier, mudFX.getDirtMultiplier);
			end;
		end;
		if #wheels > mudFX.maxWheelLimit then
			self.mud_FX.tooManyWheelsForMud = true;
			for _, wheel in pairs(wheels) do
				wheel.mudFX_sinkOnly = true;
			end;
			return;
		else
			for index, wheel in pairs(wheels) do
				local wheelLinkNode = wheel.driveNode;
				if wheelLinkNode == nil then wheelLinkNode = wheel.repr; end;
				wheel.mudFX_isCrawler = mudFX:getIsCrawler(self, wheel, crawlers);
				local hasIssue, fix = mudFX:getHasIssue(self, index, wheel);
				wheel.mudFX_sinkOnly = hasIssue;
				if not wheel.mudFX_sinkOnly then
					local mud_psData = {};
					local scaleFactor = 1.0;
					wheel.mud_psF = {};
					wheel.mud_psR = {};
					wheel.mudFXisActive = false;
					wheel.mudFXMainNode = createTransformGroup("mudNode");
					link(wheel.node, wheel.mudFXMainNode);
					local rDX1,rDY1,rDZ1 = localToWorld(self.components[1].node, 0,0,1);
					local rDX2,rDY2,rDZ2 = localToWorld(self.components[1].node, 0,1,0);
					setWorldTranslation(wheel.mudFXMainNode, unpack({localToWorld(wheelLinkNode,0,-wheel.radius,0)}));
					setDirection(wheel.mudFXMainNode,rDX1,rDY1,rDZ1,rDX2,rDY2,rDZ2);
					if fix == 1 then local bx,by,bz = getTranslation(wheels[index-1].mudFXMainNode);
					setTranslation(wheel.mudFXMainNode, -bx,by,bz);	end;
					mud_psData.psFile = "particleSystems/mud.i3d";
					mud_psData.posX = 0;--wheel.positionX;
					mud_psData.posY = 0;--wheel.positionY-wheel.radius;
					mud_psData.posZ = -(wheel.radius*1.2);--wheel.positionZ-(wheel.radius*1.2);	
					mud_psData.worldSpace = false;
					ParticleUtil.loadParticleSystemFromData(mud_psData, wheel.mud_psF, nil, false, nil, mudFX.dir, wheel.mudFXMainNode);
					mud_psData.posZ = (wheel.radius*1.2);--wheel.positionZ+(wheel.radius*1.2);
					if wheel.mudFX_isCrawler then 
						if shortCrawlers then mud_psData.posZ = crawlers[1].trackWidth*1.6;
						else mud_psData.posZ = crawlers[1].trackWidth*3.2; end;
					end;
					ParticleUtil.loadParticleSystemFromData(mud_psData, wheel.mud_psR, nil, false, nil, mudFX.dir, wheel.mudFXMainNode);
					local rX,rY,rZ = getRotation(wheel.mud_psR.shape);
					setRotation(wheel.mud_psR.shape, rX,rY-math.rad(180),rZ);
					setScale(wheel.mud_psF.shape, wheel.width * scaleFactor, wheel.radius * scaleFactor, wheel.radius * scaleFactor);
					setScale(wheel.mud_psR.shape, wheel.width * scaleFactor, wheel.radius * scaleFactor, wheel.radius * scaleFactor);
					wheel.mud_psF.resetPos = {getTranslation(wheel.mud_psF.shape)}
					wheel.mud_psR.resetPos = {getTranslation(wheel.mud_psR.shape)}
					wheel.mudFXtrackIndex = g_currentMission.tireTrackSystem:createTrack(wheel.width*2, 4);
					mudFX:loadMudPuddleToWheel(wheel, wheel.mudFXMainNode, false);
					local ir = 1;
					local il = 1;
					if wheel.additionalWheels ~= nil and #wheel.additionalWheels > 0 then
						for _,additionalWheel in pairs (wheel.additionalWheels) do
							additionalWheel.mudFX_isCrawler = false;
							if additionalWheel.tireType == WheelsUtil.getTireType("crawler") then
								additionalWheel.mudFX_isCrawler = true;
							end;
							local offset = 0.11*wheel.radius;
							if wheel.xOffset ~= nil and wheel.xOffset ~= 0 then offset = wheel.xOffset; end;
							if additionalWheel.isLeft then
								mud_psData.posX = ((wheel.width*il) + offset);--mud_psData.posX = wheel.positionX + ((wheel.width*il) + offset);
								il = il + 1;
							else
								mud_psData.posX = -((wheel.width*ir) + offset);--wheel.positionX - ((wheel.width*ir) + offset);
								ir = ir + 1;
							end;
							additionalWheel.mud_psF = {};
							additionalWheel.mud_psR = {};
							mud_psData.posZ = -(wheel.radius*1.2);--wheel.positionZ-(wheel.radius*1.2);
							ParticleUtil.loadParticleSystemFromData(mud_psData, additionalWheel.mud_psF, nil, false, nil, mudFX.dir, wheel.mudFXMainNode);
							mud_psData.posZ = (wheel.radius*1.2);--wheel.positionZ+(wheel.radius*1.2);
							if additionalWheel.mudFX_isCrawler then 
								if shortCrawlers then mud_psData.posZ = crawlers[1].trackWidth*1.6;
								else mud_psData.posZ = crawlers[1].trackWidth*3.2; end;
							end;
							ParticleUtil.loadParticleSystemFromData(mud_psData, additionalWheel.mud_psR, nil, false, nil, mudFX.dir, wheel.mudFXMainNode);
							local rX,rY,rZ = getRotation(additionalWheel.mud_psR.shape);
							setRotation(additionalWheel.mud_psR.shape, rX,rY-math.rad(180),rZ);
							setScale(additionalWheel.mud_psF.shape, additionalWheel.width * scaleFactor, additionalWheel.radius * scaleFactor, additionalWheel.radius * scaleFactor);
							setScale(additionalWheel.mud_psR.shape, additionalWheel.width * scaleFactor, additionalWheel.radius * scaleFactor, additionalWheel.radius * scaleFactor);
							additionalWheel.mud_psF.resetPos = {getTranslation(additionalWheel.mud_psF.shape)}
							additionalWheel.mud_psR.resetPos = {getTranslation(additionalWheel.mud_psR.shape)}
							additionalWheel.mudFXtrackIndex = g_currentMission.tireTrackSystem:createTrack(additionalWheel.width*2, 4);
							mudFX:loadMudPuddleToWheel(additionalWheel, wheel.mudFXMainNode, true);
						end;
					end;
				end;
			end;
		end;
		if mudFX.printVehiclesToLog and self.mud_FX ~= nil then print("Load mod: '"..mudFX.title.."' successfully applied to "..FSBaseMission:getVehicleName(self)); end;
	end;
end;

function mudFX:loadMudPuddleToWheel(wheel, linkNode, isAdditionalWheel)
	wheel.mudFXPuddle = {};
	wheel.mudFXPuddle.node = loadI3DFile(mudFX.dir.."particleSystems/mudPuddle.i3d", false, false, false);
	wheel.mudFXPuddle.node = getChildAt(wheel.mudFXPuddle.node, 0);
	wheel.mudFXPuddle.pivot = getChildAt(wheel.mudFXPuddle.node, 0);
	wheel.mudFXPuddle.start = getChildAt(wheel.mudFXPuddle.node, 1);
	wheel.mudFXPuddle.length = getChildAt(wheel.mudFXPuddle.node, 2);
	wheel.mudFXPuddle.mudShape = getChildAt(wheel.mudFXPuddle.pivot, 1);
	wheel.mudFXPuddle.waterShape = getChildAt(wheel.mudFXPuddle.pivot, 0);
	wheel.mudFXPuddle.waterShape = getChildAt(wheel.mudFXPuddle.waterShape, 0);
	if not isAdditionalWheel then wheel.mudFXPuddle.mudScroll = 0; end;
	link(linkNode, wheel.mudFXPuddle.node);
	local y = wheel.radius+0.25;
	local xx,_,_ = getWorldTranslation(wheel.mud_psF.shape);
	local _,yy,zz = getWorldTranslation(wheel.mudFXPuddle.node);
	--local gY = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, xx,yy,zz);
	--if wheel.contact ~= Wheels.WHEEL_NO_CONTACT and yy > (gY+0.5) then yy = gY; end;
	setWorldTranslation(wheel.mudFXPuddle.node, xx,yy,zz);	
	local x,_,z = getTranslation(wheel.mudFXPuddle.node);--offsetWheel,-(wheel.radius+0.25),0;
	setTranslation(wheel.mudFXPuddle.node,x,-y,z);
	wheel.mudFXPuddle.resetPos = {getTranslation(wheel.mudFXPuddle.node)};
	setScale(wheel.mudFXPuddle.node, wheel.width*6.66, wheel.radius*3.75, wheel.radius*4.25);
	setVisibility(wheel.mudFXPuddle.node, false);
end;

function mudFX:getIsCrawler(self, wheel, crawlers)	
	if wheel.wheelTire == nil then
		if self.mud_FX.crawlers ~= nil then
			if wheel.tireType == WheelsUtil.getTireType("crawler") then 
				return true; 
			else
				for _,crawler in pairs(crawlers) do
					if crawler.wheel == wheel then 
						return true; 
					end;
				end;
			end;
		end;
	end;
	return false;
end;

function mudFX:getHasIssue(self, index, wheel)
	if not wheel.hasParticles and wheel.wheelTire == nil then 
		return true, 0; 
	end;
	---------------
	if self.spec_plow ~= nil or self.spec_cultivator ~= nil or self.spec_roller then 
		return true, 0; 
	end;
	---------------
	local namesToFind = {"jcb fastrac 4220"}; --wheel[4] has local offset from rotation
	local vName = FSBaseMission:getVehicleName(self);
	vName = string.lower(vName);
	for _, nameToFind in pairs(namesToFind) do
		if string.find(vName,nameToFind) and index > 3 then 
			return false, 1;
		end;
	end;
	---------------
	--local x,y,z = getRotation(wheel.driveNode);
	--if x+y+z ~= 0 then return true, 0; end;
	---------------
	local storeCat = g_storeManager:getItemByXMLFilename(self.configFileName);
	if storeCat ~= nil and storeCat.categoryName ~= nil then
		storeCat = storeCat.categoryName;
		for _, type in pairs({ "PLANTERS", "SEEDERS", "WEEDERS", "TEDDERS",
		"WINDROWERS", "PLOWS", "SUBSOILERS", "CULTIVATORS", "DISCHARROWS", "POWERHARROWS" }) do
			if type == storeCat then 
				return true, 0;				
			end;
		end;
	end;
	---------------
	return false, 0;
end;

function mudFX:getIsGroundWet()
	local rainWeather = false;
	if mudFX.overrideRain then 
		rainWeather = true;
	else
		local rainScale = g_currentMission.environment.weather:getRainFallScale()
		local timeSinceLastRain = g_currentMission.environment.weather:getTimeSinceLastRain()
		local wetGround = mudFX.useWetness and (g_currentMission.environment.groundWetness >= 0.01);
		if wetGround then
			rainWeather = true;
		elseif not (rainScale <= 0.1 and timeSinceLastRain > 20) then
			rainWeather = true;
		end;
	end;
	return rainWeather;
end;

function mudFX:getIsTireOnDirt(wheel,baseWheel)
	local isOnDirt = false;
	if g_currentMission.terrainRootNode ~= nil and g_currentMission.terrainDetailId ~= nil then
		local wheelNode = wheel.wheelTire;
		if wheelNode == nil then wheelNode = wheel.driveNode; end; --crawler
		if wheelNode == nil then wheelNode = wheel.repr; end;
		local wheelX, wheelY, wheelZ = getWorldTranslation(wheelNode);
		wheelY = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, wheelX, wheelY, wheelZ);
		local bits = getDensityAtWorldPos(g_currentMission.terrainDetailId, wheelX, wheelY, wheelZ);--wheel.densityBits
		--[[if baseWheel == nil then baseWheel = wheel; end;
		if baseWheel.lastTerrainValue == 0 then bits = 0 end;]] --Stopped certain vehicles from detecting fields..
		if bits ~= 0 then--and bits ~= 256  and bits ~= g_currentMission.grassValue then --is on a field
			isOnDirt = true;
		--elseif mudFX:getIsTireOnMud(wheel,FillType.DIRT) then
		--	isOnDirt = true;
		elseif mudFX.useTerrainColors then
			local cR,cG,cB,_,_ = getTerrainAttributesAtWorldPos(g_currentMission.terrainRootNode, wheelX, wheelY, wheelZ, true, true, true, true, false);
			if (cR > 0.15500 and cR < 0.15501) and (cG > 0.08200 and cG < 0.08201) and (cB > 0.03700 and cB < 0.03701) then -- is on dirt terrain			
				isOnDirt = true;
			end;
			if (cR > 0.36 and cR < 0.36001) and (cG > 0.12499 and cG < 0.12501) and (cB > 0.056 and cB < 0.05601) then --is on red dirt terrain
				isOnDirt = true;
			end;
		end;
	end;
	return isOnDirt;
end;

function mudFX:getIsTireOnMud(wheel, filltype)
	if g_currentMission.terrainRootNode ~= nil and g_currentMission.terrainDetailId ~= nil then
		if filltype ~= nil then
			if filltype == 0 then filltype = FillType.MUD end;
			local wheelNode = wheel.wheelTire;
			if wheelNode == nil then wheelNode = wheel.driveNode; end; --crawler
			if wheelNode == nil then wheelNode = wheel.repr; end;
			local x1,y1,z1 = localToWorld(wheelNode, -wheel.width/2,-wheel.radius,0);
			local x2,y2,z2 = localToWorld(wheelNode, wheel.width/2,-wheel.radius,0);
			y1 = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, x1,y1,z1);
			y2 = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, x2,y2,z2);
			local foundFillType = DensityMapHeightUtil.getFillTypeAtLine(x1,y1,z1, x2,y2,z2, 1.0);
			if foundFillType == filltype then
				return true;
			end;
		end;
	end;
	return false;
end;

function mudFX:stopPS(wheel)
	if wheel.mud_psF ~= nil then ParticleUtil.setEmittingState(wheel.mud_psF, false, false, false); end;
	if wheel.mud_psR ~= nil then ParticleUtil.setEmittingState(wheel.mud_psR, false, false, false); end;
	wheel.mudFXisActive = false;
end;

function mudFX:stopEveryPS(self)
	if self.mud_FX ~= nil then
		local wheels = self.spec_wheels.wheels;	
		if wheels ~= nil and #wheels > 0 and not self.mud_FX.tooManyWheelsForMud then
			for w=1, #wheels do
				local wheel = wheels[w];
				if wheel.mudFXisActive ~= nil then
					mudFX:stopPS(wheel);			
					if wheel.additionalWheels ~= nil and #wheel.additionalWheels > 0 then
						for _,additionalWheel in pairs (wheel.additionalWheels) do
							mudFX:stopPS(additionalWheel);
						end;
					end;
				end;
			end;
		end;
	end;
end;

function mudFX:scrollMudUVs(self, wheel, dt)
	local dir = self.movingDirection;
	if dir ~= 0 then
		local wheelSpeed = MathUtil.rpmToMps(wheel.netInfo.xDriveSpeed / (2*math.pi) * 60, wheel.radius);
		local scrollSpeed = wheelSpeed*3;--(mudFX.uvSpeed*dt);
		local position = wheel.mudFXPuddle.mudScroll;
		position = position+scrollSpeed;
		if position >= 200 then position = 0 end;
		setShaderParameter(wheel.mudFXPuddle.mudShape, "offsetUV", 0, position, 0, 0, false);
		local sVals = {getShaderParameter(wheel.mudFXPuddle.waterShape, "bumpSpeedScale")};
		setShaderParameter(wheel.mudFXPuddle.waterShape, "bumpSpeedScale", sVals[1], sVals[2], sVals[3]-(wheelSpeed*-1), sVals[4], false);
		wheel.mudFXPuddle.mudScroll = position;
		if wheel.additionalWheels ~= nil and #wheel.additionalWheels > 0 then
			for _, aWheel in pairs(wheel.additionalWheels) do
				setShaderParameter(aWheel.mudFXPuddle.mudShape, "offsetUV", 0, position, 0, 0, false);
				setShaderParameter(aWheel.mudFXPuddle.waterShape, "bumpSpeedScale", sVals[1], sVals[2], sVals[3]-(wheelSpeed*-1), sVals[4], false);
			end;
		end;
	end;
end;

function mudFX:correctPSLocation(wheel)	
	local _,y,_ = getTranslation(wheel.mud_psF.shape);
	local _,y2,_ = getTranslation(wheel.mud_psR.shape);
	setTranslation(wheel.mud_psF.shape, wheel.mud_psF.resetPos[1], y, wheel.mud_psF.resetPos[3]);
	setTranslation(wheel.mud_psR.shape, wheel.mud_psR.resetPos[1], y2, wheel.mud_psR.resetPos[3]);
end;

function mudFX:correctPuddleLocation(wheel)
	local puddle = wheel.mudFXPuddle;
	local x,y,z = getWorldTranslation(puddle.node);
	local xF,yF,zF = getWorldTranslation(puddle.start);
	local xR,yR,zR = getWorldTranslation(puddle.length);
    y = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, x,y,z);
    y1 = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, xF,yF,zF);
    y2 = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, xR,yR,zR);	
	if yF > y1 then
		if (y2 <= (y+.01) and y2 >= (y-0.01)) and y1 < y then
			setWorldTranslation(puddle.node,x,y1,z);
		end;
	elseif yR > y2 then	
		if (y1 <= (y+.01) and y1 >= (y-0.01)) and y2 < y then
			setWorldTranslation(puddle.node,x,y2,z);
		end;
	end;
	local _,y,_ = getTranslation(puddle.node); -- finalize position using local space with stored x and z values.
	setTranslation(puddle.node, puddle.resetPos[1], y, puddle.resetPos[3]);

	--[[local _,dY,_ = localToWorld(wheel.driveNode,0,-wheel.radius,0);
	if dY > y and dY > y1 and dY > y2 then
		if wheel.contact ~= nil and wheel.contact == Wheels.WHEEL_NO_CONTACT then
			mudFX:hidePuddle(wheel);
		end;
	end;]]
end;

function mudFX:setPSHeight(wheel)
	local x,y,z = getWorldTranslation(wheel.mudFXMainNode);
	local gY = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, x,y,z);
	--if y < gY then
		local xT,_,zT = getWorldTranslation(wheel.mud_psF.shape);
		local xT2,_,zT2 = getWorldTranslation(wheel.mud_psR.shape);
		local xT3,_,zT3 = getWorldTranslation(wheel.mudFXPuddle.node);
		setWorldTranslation(wheel.mud_psF.shape, xT,gY,zT);
		setWorldTranslation(wheel.mud_psR.shape, xT2,gY,zT2);
		setWorldTranslation(wheel.mudFXPuddle.node, xT3,gY,zT3);
		mudFX:correctPSLocation(wheel);
		--mudFX:correctPuddleLocation(wheel);
		if wheel.additionalWheels ~= nil and #wheel.additionalWheels > 0 then
			for _, aWheel in pairs(wheel.additionalWheels) do
				local xx,_,zz = getWorldTranslation(aWheel.mud_psF.shape);	
				local xx2,_,zz2 = getWorldTranslation(aWheel.mud_psR.shape);
				local xx3,yy3,zz3 = getWorldTranslation(aWheel.mudFXPuddle.node);
				gY = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, xx3,yy3,zz3);
				setWorldTranslation(aWheel.mud_psF.shape, xx,gY,zz);
				setWorldTranslation(aWheel.mud_psR.shape, xx2,gY,zz2);	
				setWorldTranslation(aWheel.mudFXPuddle.node, xx3,gY,zz3);
				mudFX:correctPSLocation(aWheel);	
				--mudFX:correctPuddleLocation(aWheel);
			end;
		end;
	--[[else
		local _,yT,_ = getTranslation(wheel.mud_psF.shape);
		local _,yTR,_ = getTranslation(wheel.mud_psR.shape);
		if yT ~= (-wheel.radius) or yTR ~= (-wheel.radius) then
			setTranslation(wheel.mud_psF.shape, unpack(wheel.mud_psF.resetPos));
			setTranslation(wheel.mud_psR.shape, unpack(wheel.mud_psR.resetPos));
			setTranslation(wheel.mudFXPuddle.node, unpack(wheel.mudFXPuddle.resetPos));
		end;
		if wheel.additionalWheels ~= nil and #wheel.additionalWheels > 0 then
			for _, aWheel in pairs(wheel.additionalWheels) do
				local _,yT2,_ = getTranslation(aWheel.mud_psF.shape);	
				local _,yTR2,_ = getTranslation(aWheel.mud_psR.shape);
				if yT2 ~= (-wheel.radius) or yTR2 ~= (-wheel.radius) then
					setTranslation(aWheel.mud_psF.shape, unpack(aWheel.mud_psF.resetPos));
					setTranslation(aWheel.mud_psR.shape, unpack(aWheel.mud_psR.resetPos));
					setTranslation(aWheel.mudFXPuddle.node, unpack(aWheel.mudFXPuddle.resetPos));
				end;
			end;
		end;
	end;]]
end;

function mudFX:updateMudResistance(self, wheel, dt)
	local isAttachment = false;
	local dividend = 1;
	if self.spec_attacherJoints ~= nil then
		dividend = 1+#self.spec_attacherJoints.attachedImplements;
	end;
	if self.spec_attachable ~= nil and self.spec_attachable.attacherVehicle ~= nil then
		if self ~= self.spec_attachable.attacherVehicle:getRootVehicle() then
			isAttachment = true;
		end;
	end;
	local multiplier = 2.5;--*1.5
	if wheel.tireType ~= WheelsUtil.getTireType("mud") then
		if wheel.tireType ~= WheelsUtil.getTireType("crawler") then
			multiplier = multiplier*1.2;
		end;
	end;
	multiplier = multiplier/dividend;
	if isAttachment then multiplier = multiplier/2; end;
	local brakeFactor = mudFX.mudDensity*multiplier-((wheel.width*2)*0.1);
	if brakeFactor < 0.1 then brakeFactor = 0.1 end;
	if wheel.isCareWheel then
		brakeFactor = brakeFactor*1.25;
	end;
	if wheel.additionalWheels ~= nil then
		brakeFactor = brakeFactor-((brakeFactor/#wheel.additionalWheels)/2);
	end;
	if not isAttachment then
		local tspeed = self:getLastSpeed(true);
		local rootMotor = self:getMotor();
		local vTorq = 1;
		if rootMotor ~= nil then vTorq = math.max(rootMotor:getPeakTorque()/1.75,0.1); end;
		local limiter = 30*vTorq;
		if tspeed > limiter then  -- Splash Mud Forwards
			tspeed = (tspeed/10)-3;
			tspeed = tspeed*10;
			brakeFactor=brakeFactor*tspeed;
			--enable Splash PS?
			--self.mud_FX.mudSplash = true;
		end;
	end;
	if brakeFactor > 0 then
		WheelsUtil.updateWheelPhysics(self, wheel, brakeFactor, dt);
	end;
end;

function mudFX:updateWheelRotation(self, wheel)
	if self.mud_FX.crawlers == nil or wheel.wheelTire ~= nil then
		if wheel.mudFXPuddle ~= nil then
			if getVisibility(wheel.mudFXPuddle.node) or wheel.mudFXisActive then
				--local _,y,_ = getRotation(wheel.repr);
				setRotation(wheel.mudFXMainNode,0,wheel.steeringAngle,0);
			end;		
		end;		
	end;
end;
					
function mudFX:updatePuddles(self, dt)
	for _, wheel in pairs(self.spec_wheels.wheels) do
		if wheel.mudFXPuddle ~= nil then		
			if getVisibility(wheel.mudFXPuddle.node) then
				mudFX:correctPuddleLocation(wheel);
			end;			
			if wheel.additionalWheels ~= nil and #wheel.additionalWheels > 0 then
				for _, aWheel in pairs(wheel.additionalWheels) do
					if aWheel.mudFXPuddle ~= nil then
						if getVisibility(aWheel.mudFXPuddle.node) then
							mudFX:correctPuddleLocation(aWheel);
						end;
					end;
				end;
			end;
		end;
	end;
end;

function mudFX:updatePS(self, dt)
	local speedfactor = mudFX.psSpeedRatio;
	local enteredSpec = self.spec_enterable;
	local attachableSpec = self.spec_attachable;
	local isEntered = false;
	local isAttachment = false;
	if enteredSpec ~= nil then isEntered = Utils.getNoNil(enteredSpec.isEntered, false); end;
	if not isEntered and attachableSpec ~= nil and attachableSpec.attacherVehicle ~= nil then
		local attachedVehicle = attachableSpec.attacherVehicle:getRootVehicle();
		if attachedVehicle.spec_enterable ~= nil then
			isEntered = Utils.getNoNil(attachedVehicle.spec_enterable.isEntered, false);
			isAttachment = true;
		end;
	end;
	if isEntered and not self.tooManyWheelsForMud then --Wetground & not an excluded attachment		
		--if self.movingDirection ~= 0 then
			local wheels = self.spec_wheels.wheels;
			for w=1, #wheels do
				local wheel = wheels[w];
				if wheel.mudFXisActive ~= nil then
					local tspeed = 0;
					if isAttachment then tspeed = self:getLastSpeed(true);
					elseif wheel.netInfo.xDriveSpeed ~= nil then 
						tspeed = MathUtil.rpmToMps(wheel.netInfo.xDriveSpeed/(2*math.pi)*60,wheel.radius)*3600;--self:getLastSpeed(true);
						if tspeed < 0 then tspeed = tspeed*-1; end;
					end;
					if not wheel.mudFXisActive then   --[[Activate]]
						--print(tspeed.."  :  "..self:getLastSpeed(true))
						if tspeed > 2.5 and ((mudFX:getIsGroundWet() and mudFX:getIsTireOnDirt(wheel)) or mudFX:getIsTireOnMud(wheel,0)) then							
							if wheel.contact ~= Wheels.WHEEL_NO_CONTACT then wheel.mudFXisActive = true; end;
						end;
					else
						mudFX:setPSHeight(wheel);
						mudFX:updateSink(wheel, true);
						mudFX:scrollMudUVs(self, wheel, dt);
						mudFX:updateMudResistance(self, wheel, dt);
						if tspeed <= 2.5 then----------------------------
							mudFX:stopPS(wheel);
							mudFX:setPSHeight(wheel);
							mudFX:updateSink(wheel, false);
						elseif self.movingDirection < 0 then
						----------------------------Update Reverse--------------------------------------------------
							ParticleUtil.setEmittingState(wheel.mud_psF, false);
							ParticleUtil.setEmittingState(wheel.mud_psR, true);
							ParticleUtil.setParticleSystemSpeed(wheel.mud_psR, tspeed/(speedfactor/(wheel.radius*2)));	
						elseif self.movingDirection > 0 then
						----------------------------Update Forwards--------------------------------------------------
							ParticleUtil.setEmittingState(wheel.mud_psR, false);				
							ParticleUtil.setEmittingState(wheel.mud_psF, true);
							ParticleUtil.setParticleSystemSpeed(wheel.mud_psF, tspeed/(speedfactor/(wheel.radius*2)));						
						--[[else							
							local psDirection = Utils.getNoNil(self.spec_motorized.lastDirection,1);							
							ParticleUtil.setEmittingState(wheel.mud_psF, (psDirection==1));				
							ParticleUtil.setEmittingState(wheel.mud_psR, (psDirection==(-1)));
							ParticleUtil.setParticleSystemSpeed(wheel.mud_psF, tspeed/(speedfactor/wheel.radius));]]
						end;----------------------------
					end;
					if (not mudFX:getIsGroundWet() or not mudFX:getIsTireOnDirt(wheel)) and not mudFX:getIsTireOnMud(wheel,0) then
						mudFX:stopPS(wheel);
						mudFX:setPSHeight(wheel);
						mudFX:updateSink(wheel, false);
						mudFX:hidePuddle(wheel);
					end;
					------------------------------------------------------------------
					if wheel.additionalWheels ~= nil and #wheel.additionalWheels > 0 then
						for _,additionalWheel in pairs (wheel.additionalWheels) do
							if not additionalWheel.mudFXisActive then
								if tspeed > 2.5 and ((mudFX:getIsGroundWet() and mudFX:getIsTireOnDirt(additionalWheel,wheel)) or mudFX:getIsTireOnMud(additionalWheel,0)) then
									additionalWheel.mudFXisActive = true;
								end;
							else
								if tspeed <= 2.5 then----------------------------								
									mudFX:stopPS(additionalWheel);
								elseif self.movingDirection < 0 then
								----------------------------Update Reverse--------------------------------------------------
									ParticleUtil.setEmittingState(additionalWheel.mud_psF, false);
									ParticleUtil.setEmittingState(additionalWheel.mud_psR, true);
									ParticleUtil.setParticleSystemSpeed(additionalWheel.mud_psR, tspeed/(speedfactor/(additionalWheel.radius*2)));	
								elseif self.movingDirection > 0 then
								----------------------------Update Forwards--------------------------------------------------
									ParticleUtil.setEmittingState(additionalWheel.mud_psR, false);
									ParticleUtil.setEmittingState(additionalWheel.mud_psF, true);
									ParticleUtil.setParticleSystemSpeed(additionalWheel.mud_psF, tspeed/(speedfactor/(additionalWheel.radius*2)));
								end;----------------------------
							end;
							if (not mudFX:getIsGroundWet() or not mudFX:getIsTireOnDirt(additionalWheel,wheel)) and not mudFX:getIsTireOnMud(additionalWheel,0) then 
								mudFX:stopPS(additionalWheel);
								mudFX:hidePuddle(additionalWheel);
							end;
						end;
					end;
					----------------------------------------------------------------------
					mudFX:updateWheelRotation(self, wheel);
				elseif wheel.mudFX_sinkOnly then
					if (mudFX:getIsGroundWet() and mudFX:getIsTireOnDirt(wheel)) or mudFX:getIsTireOnMud(wheel,0) then
						-- the following method will make tracks sink only if both ends are in mud. *may be needed to prevent sliding of still tracks.
						--[[local sinkSimWheel = false;
						for _,otherWheel in pairs(self.spec_wheels.wheels) do
							if otherWheel.isLeft == wheel.isLeft then
								if otherWheel.sinkInMud ~= nil and otherWheel.sinkInMud then 
									sinkSimWheel = true;
								end;
							end;
						end;
						mudFX:updateSink(wheel, sinkSimWheel);]]
						mudFX:updateSink(wheel, true);
					else
						mudFX:updateSink(wheel, false);
					end;
				end;
			end;
	elseif not self.tooManyWheelsForMud then
		mudFX:stopEveryPS(self);
		for _,wheel in pairs (self.spec_wheels.wheels) do
			if wheel.mudFXPuddle ~= nil or (wheel.mudFX_sinkOnly ~= nil and wheel.mudFX_sinkOnly) then
				mudFX:updateSink(wheel, false);
				--[[mudFX:hidePuddle(wheel);
				if wheel.additionalWheels ~= nil and #wheel.additionalWheels > 0 then
					for _,additionalWheel in pairs (wheel.additionalWheels) do
						mudFX:hidePuddle(additionalWheel);
					end;
				end;]]
			end;
		end;
	end;

end;

function mudFX:updateSink(wheel, activated) --Also the Puddle Visibility Controller
	if activated then
		if wheel.sinkInMud == nil then 
			wheel.sinkInMud = true;--Activates wheel sinking				
			if not wheel.mudFX_sinkOnly and not wheel.mudFX_isCrawler then
				local x,y,z = getWorldTranslation(wheel.mudFXMainNode);
				gY = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, x,y,z)+0.1;
				if y < gY then --Must be lower than ground+0.1m
					if not getVisibility(wheel.mudFXPuddle.node) then
						setVisibility(wheel.mudFXPuddle.node, true);
					end;
				else
					mudFX:hidePuddle(wheel);
				end;
				if wheel.additionalWheels ~= nil and #wheel.additionalWheels > 0 then
					for _, additionalWheel in pairs(wheel.additionalWheels) do					
						if additionalWheel.mudFXPuddle.node ~= nil then
							if y < gY then
								if not getVisibility(additionalWheel.mudFXPuddle.node) then
									setVisibility(additionalWheel.mudFXPuddle.node, true);
								end;
							else
								mudFX:hidePuddle(additionalWheel);
							end;
						end;
					end;
				end;
			end;
		end;
	else
		if wheel.sinkInMud ~= nil then
			wheel.sinkInMud = nil;--Stops wheel sinking	
		end;
	end;	
end;

function mudFX:hidePuddle(wheel)
	if getVisibility(wheel.mudFXPuddle.node) then
		setVisibility(wheel.mudFXPuddle.node, false);
		setTranslation(wheel.mudFXPuddle.node, unpack(wheel.mudFXPuddle.resetPos));
	end;
end;

function mudFX:createBodyTrack(self,index,r,g,b,a,ux,uy,uz,dirt,dir)
	if self.spec_attachable == nil then
		local wheels = self.spec_wheels.wheels;
		if self.mudFXMudTrackIndex == nil and #wheels > 1 then
			local x1,y1,z1 = getWorldTranslation(wheels[1].driveNode);
			local oppositeWheelIndex = wheels[1].oppositeWheelIndex;
			if oppositeWheelIndex == nil then oppositeWheelIndex = 2; end;
			if wheels[oppositeWheelIndex] ~= nil and wheels[oppositeWheelIndex].isLeft == wheels[1].isLeft then
				oppositeWheelIndex = 3; 
			end;
			if wheels[oppositeWheelIndex] ~= nil and wheels[oppositeWheelIndex].isLeft ~= wheels[1].isLeft then
				local x2,y2,z2 = getWorldTranslation(wheels[oppositeWheelIndex].driveNode);
				local width = x1-x2;
				local width2 = y1-y2;
				local width3 = z1-z2;
				if width < 0 then width = width * (-1); end;
				if width2 < 0 then width2 = width2 * (-1); end;
				if width3 < 0 then width3 = width3 * (-1); end;
				width = (width + width2) + width3;
				width = (width - wheels[1].width);
				self.mudFXMudTrackIndex = g_currentMission.tireTrackSystem:createTrack(width*0.8, 4);
			end;
		end;
		if self.mudFXMudTrackIndex ~= nil then
			local x,y,z = getWorldTranslation(self.components[1].node);
			y = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, x,y,z)+0.001;
			g_currentMission.tireTrackSystem:addTrackPoint(self.mudFXMudTrackIndex,x,y,z,ux,uy,uz,r,g,b,dirt,a,dir);
		end;
	end;
end;

function mudFX:onUpdate(dt)
	if self.spec_wheels ~= nil then
		if self.spec_wheels.wheels ~= nil and #self.spec_wheels.wheels > 0 then
			mudFX:updatePS(self, dt);
			mudFX:updatePuddles(self, dt);
			--[[if self.mud_FX ~= nil and not self.mud_FX.initialized then 
				for _, wheel in pairs(self.spec_wheels.wheels) do
					if wheel.mudFXMainNode ~= nil and wheel.wheelTire ~= nil then
						local xTire,yTire,zTire = localToWorld(wheel.wheelTire,0,-wheel.radius,0);
						if xTire ~= x and yTire ~= y and zTire ~= z then
							setWorldTranslation(wheel.mudFXMainNode, unpack({localToWorld(wheel.wheelTire,0,-wheel.radius,0)}));
						end;
					end;
				end;
				self.mud_FX.initialized = true;
			end;]]
		end;
	end;
end;

function mudFX:onLeave()
	mudFX:stopEveryPS(self);
end;

function mudFX:getDirtMultiplier()
    local spec = self.spec_washable;
    local multiplier = 1;
    if self:getLastSpeed() < 1 then
        multiplier = 0;
    end;
    if self:getIsOnField() then
        multiplier = multiplier * spec.fieldMultiplier;
        local wetness = g_currentMission.environment.weather:getGroundWetness();
        if wetness > 0 then
            multiplier = multiplier * (1 + wetness);
        end;
	end;
	local isInMud = false;
	if self.spec_wheels ~= nil and self.spec_wheels.wheels ~= nil then
		for _, wheel in pairs(self.spec_wheels.wheels) do
			if ((mudFX:getIsGroundWet() and mudFX:getIsTireOnDirt(wheel)) or mudFX:getIsTireOnMud(wheel,0)) then
				isInMud = true;
				break;
			end;
		end;
	end;
	if isInMud then
		if self.spec_washable ~= nil and self.spec_washable.washableNodes ~= nil then
			for _, nodeData in ipairs(self.spec_washable.washableNodes) do
				if self:getNodeDirtAmount(nodeData) < 1 then
					self:setNodeDirtAmount(nodeData, self:getNodeDirtAmount(nodeData)+0.001, true);
				end;
			end;
		end;
		--[[if self.getAttachedImplements ~= nil then
			local attachedImplements = self:getAttachedImplements();
			if attachedImplements ~= nil then
				for _, implement in pairs(attachedImplements) do
					if implement.object ~= nil and implement.object.spec_wheels == nil then
						if implement.object.spec_washable ~= nil and implement.object.spec_washable.washableNodes ~= nil then
							for _, nodeData in ipairs(implement.object.spec_washable.washableNodes) do
								implement.object:setNodeDirtAmount(nodeData, implement.object:getNodeDirtAmount(nodeData)+0.001, true);
							end;
						end;
					end;
				end;
			end;
		end;]]
		multiplier = multiplier * (40.0*Washable.getIntervalMultiplier()); 
	end;
    return multiplier;
end;

print("Load mod: "..mudFX.title.." : v"..mudFX.version.." by "..mudFX.author);