----------------------------------------------------------------------------
-- @Author: ViperGTS96------------------------------------------------------
----------------------------------------------------------------------------
--------------------"The simplest design is the best design." --------------
----------------------------------------------------------------------------
----------------------------------------------------------------------------

Wheels.updateWheelSink = function (self, wheel, dt)
    if wheel.supportsWheelSink then
        if self.isServer and self.isAddedToPhysics then
            local spec = self.spec_wheels;
            -- map noise to an asbolute value or to a certain percentage of the wheel radius?
            local maxSink = 0.20;
            local sinkTarget = 0;
            if wheel.mirroredWheel == nil then
                for _, mirWheel in ipairs(spec.wheels) do
                    if mirWheel.mirroredWheel == nil and mirWheel ~= wheel then -- only the first wheel got the mirrored one
                        local x1, y1, z1 = localToLocal(wheel.node, wheel.repr, 0, 0, 0);
                        local x2, y2, z2 = localToLocal(wheel.node, mirWheel.repr, 0, 0, 0);
                        local diff = math.abs(x1-(-x2)) + math.abs(y1-y2) + math.abs(z1-z2);
                        if diff < 0.25 then
                            wheel.mirroredWheel = mirWheel;
                            mirWheel.invMirroredWheel = wheel;
                        end;
                    end;
                end;
            end;
            local force = false;
            if wheel.contact ~= Wheels.WHEEL_NO_CONTACT and self:getLastSpeed() > 0.3 then
                wheel.avgSink = nil;
                local width = 0.25 * wheel.width;
                local length = 0.25 * wheel.width;
                local x,_,z = localToLocal(wheel.driveNode, wheel.repr, 0,0,0);
                local x0,_,z0 = localToWorld(wheel.repr, x + width, 0, z - length);
                local x1,_,z1 = localToWorld(wheel.repr, x - width, 0, z - length);
                local x2,_,z2 = localToWorld(wheel.repr, x + width, 0, z + length);
                local x,z, widthX,widthZ, heightX,heightZ = MathUtil.getXZWidthAndHeight(x0, z0, x1, z1, x2, z2);
                local density, area = FSDensityMapUtil.getFieldValue(x0, z0, x1, z1, x2, z2);
                local terrainValue = 0;
                if area > 0 then
                    terrainValue = math.floor(density/area + 0.5);
                end;
                wheel.lastTerrainValue = terrainValue;
                local noiseValue = 0;
                if terrainValue > 0 then
                    local xPerlin = x + 0.5*widthX + 0.5*heightX;
                    local zPerlin = z + 0.5*widthZ + 0.5*heightZ;
                    -- Round to 1cm to avoid sliding when not moving
                    xPerlin = math.floor(xPerlin*100)*0.01;
                    zPerlin = math.floor(zPerlin*100)*0.01;
                    local perlinNoise = Wheels.perlinNoiseSink;
                    local noiseSink = 0.5 * (1 + getPerlinNoise2D(xPerlin*perlinNoise.randomFrequency, zPerlin*perlinNoise.randomFrequency, perlinNoise.persistence, perlinNoise.numOctaves, perlinNoise.randomSeed));
                    perlinNoise = Wheels.perlinNoiseWobble;
                    local noiseWobble = 0.5 * (1 + getPerlinNoise2D(xPerlin*perlinNoise.randomFrequency, zPerlin*perlinNoise.randomFrequency, perlinNoise.persistence, perlinNoise.numOctaves, perlinNoise.randomSeed));
                    -- estimiate pressure on surface
                    local gravity = 9.81;
                    local tireLoad = getWheelShapeContactForce(wheel.node, wheel.wheelShape);
                    if tireLoad ~= nil then
                        local nx,ny,nz = getWheelShapeContactNormal(wheel.node, wheel.wheelShape);
                        local dx,dy,dz = localDirectionToWorld(wheel.node, 0,-1,0);
                        tireLoad = -tireLoad*MathUtil.dotProduct(dx,dy,dz, nx,ny,nz);
                        tireLoad = tireLoad + math.max(ny*gravity, 0.0) * wheel.mass; -- add gravity force of tire
                    else
                        tireLoad = 0;
                    end;
                    tireLoad = tireLoad / gravity;
                    local loadFactor = math.min(1.0, math.max(0, tireLoad / wheel.maxLatStiffnessLoad));
                    local wetnessFactor = g_currentMission.environment.weather:getGroundWetness();
                    noiseSink = 0.333*(2*loadFactor + wetnessFactor) * noiseSink;
                    noiseValue = math.max(noiseSink, noiseWobble);
                end;
				------------------------------------------------------------------------------------------
				if wheel.sinkInMud == nil then 
					maxSink = Wheels.MAX_SINK[terrainValue] or maxSink;
				else
					local mudSinkFactor = wheel.radius+0.1;
					if wheel.isCareWheel == nil or not wheel.isCareWheel then
						local extraWheelCount = 0;
						if wheel.additionWheels ~= nil then extraWheelCount = #wheel.additionWheels; end;
						mudSinkFactor = mudSinkFactor - (mudSinkFactor*((wheel.width*0.25)*extraWheelCount));
						if wheel.mudFX_isCrawler ~= nil and wheel.mudFX_isCrawler then mudSinkFactor = mudSinkFactor/2; end;
						if mudSinkFactor <= 0 then mudSinkFactor = 0.05; end;
					end;
					maxSink = mudSinkFactor;
				end;
				------------------------------------------------------------------------------------------
                -- plowing effect
                if terrainValue == 2 and wheel.oppositeWheelIndex ~= nil then
                    local oppositeWheel = spec.wheels[wheel.oppositeWheelIndex];
                    if oppositeWheel.lastTerrainValue ~= nil and oppositeWheel.lastTerrainValue ~= 2 then
                        maxSink = maxSink * 1.3;
                    end;
                end;
				------------------------------------------------------------------------------------------
				if wheel.sinkInMud == nil then 
					sinkTarget = math.min(0.2*wheel.radiusOriginal, math.min(maxSink, wheel.maxWheelSink) * noiseValue);
				else
					sinkTarget = maxSink;
					----------
					--[[local bumpIntheMud = math.random(1,100);
					if bumpIntheMud > 90 then 	
						local dir = -1;
						if wheel.isLeft then dir=1; end;				
						local forceLvl = (math.random(9,26)*(self:getLastSpeed()/30))*(wheel.mass+wheel.width+wheel.radius);
						addForce(wheel.node,0,forceLvl,wheel.width,2*dir,(-wheel.radius)*2,0,true);
					end;]]
					------------
				end;
				------------------------------------------------------------------------------------------
            elseif self:getLastSpeed() < 0.3 then
                -- if we are standing still we try to synchronize the left and right wheel radius to avoid wobbeling
                if wheel.mirroredWheel ~= nil then
                    if wheel.avgSink == nil then
                        wheel.avgSink = (wheel.mirroredWheel.sinkTarget + wheel.sinkTarget) / 2;
                    end;
                    sinkTarget = wheel.avgSink;
                    force = wheel.sink ~= wheel.avgSink;
                    wheel.sinkTarget = sinkTarget;
                elseif wheel.invMirroredWheel ~= nil then
                    if wheel.invMirroredWheel.avgSink ~= nil then
                        sinkTarget = wheel.invMirroredWheel.avgSink;
                        force = wheel.sink ~= wheel.invMirroredWheel.avgSink;
                        wheel.sinkTarget = sinkTarget;
                    end;
                end;
            end;
            if wheel.sinkTarget < sinkTarget then
                wheel.sinkTarget = math.min(sinkTarget, wheel.sinkTarget + (0.05 * math.min(30, math.max(0, self:getLastSpeed()-0.2)) * (dt/1000)));
            elseif wheel.sinkTarget > sinkTarget then
                wheel.sinkTarget = math.max(sinkTarget, wheel.sinkTarget - (0.05 * math.min(30, math.max(0, self:getLastSpeed()-0.2)) * (dt/1000)));
            end;
            if math.abs(wheel.sink - wheel.sinkTarget) > 0.001 or force then
                wheel.sink = wheel.sinkTarget;
                local radius = wheel.radiusOriginal - wheel.sink;
                if radius ~= wheel.radius then
                    wheel.radius = radius;
                    if self.isServer then
                        self:setWheelPositionDirty(wheel);
                        local sinkFactor = (wheel.sink/maxSink) * (1 + (0.4 * g_currentMission.environment.weather:getGroundWetness()));
                        wheel.sinkLongStiffnessFactor = (1.0 - (0.10 * sinkFactor));
                        wheel.sinkLatStiffnessFactor  = (1.0 - (0.20 * sinkFactor));
                        self:setWheelTireFrictionDirty(wheel);
                    end;
                end;
            end;
        end;
    end;
end;

Wheels.updateWheelTireTracks = function (self, wheel)
    local wx, wy, wz = wheel.netInfo.x, wheel.netInfo.y, wheel.netInfo.z;
    wy = wy - wheel.radius;
    wx = wx + wheel.xOffset;
    wx, wy, wz = localToWorld(wheel.node, wx,wy,wz);
    local r, g, b, a, t = self:getTireTrackColor(wheel, wx, wy, wz);
    if wheel.tireTrackIndex ~= nil then
        if self:getAllowTireTracks() and r ~= nil then
            local ux,uy,uz = localDirectionToWorld(wheel.node, 0, 1, 0);
            local tireDirection = self.movingDirection;
            if wheel.tireIsInverted then
                tireDirection = tireDirection * -1;
            end;
			-------------------------------------------------------------------------------
			local wetGround = mudFX:getIsGroundWet();
			if wheel.mudFXtrackIndex ~= nil and ((wetGround and mudFX:getIsTireOnDirt(wheel)) or (wheel.mudFXPuddle ~= nil and getVisibility(wheel.mudFXPuddle.node))) and not mudFX:getIsTireOnMud(wheel,0) then --Fillplanes sit higher than tiretracks
				--wx, wy, wz = localToWorld(wheel.node, wheel.netInfo.x+wheel.xOffset,wheel.netInfo.y-wheel.radius,wheel.netInfo.z-wheel.radius);
				g_currentMission.tireTrackSystem:addTrackPoint(wheel.mudFXtrackIndex, wx, wy, wz, ux, uy, uz, r, g, b, wheel.dirtAmount, a, tireDirection);
				if wheel == self.spec_wheels.wheels[1] then mudFX:createBodyTrack(self, wheel.mudFXtrackIndex, r,g,b,a,ux,uy,uz, wheel.dirtAmount, tireDirection); end;
			else
			-------------------------------------------------------------------------------
				g_currentMission.tireTrackSystem:addTrackPoint(wheel.tireTrackIndex, wx, wy, wz, ux, uy, uz, r, g, b, wheel.dirtAmount, a, tireDirection);
				if wheel.mudFXtrackIndex ~= nil then g_currentMission.tireTrackSystem:cutTrack(wheel.mudFXtrackIndex); end;
			end;
            if wheel.additionalWheels ~= nil then
                for _, additionalWheel in pairs(wheel.additionalWheels) do
                    if additionalWheel.tireTrackIndex ~= nil then
                        wx, wy, wz = worldToLocal(wheel.node, getWorldTranslation(additionalWheel.wheelTire));
                        wy = wy - wheel.radius;
                        wx = wx + wheel.xOffset;
                        wx, wy, wz = localToWorld(wheel.node, wx,wy,wz);
                        wy = math.max(wy, getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, wx, wy, wz));
                        tireDirection = self.movingDirection;
                        if additionalWheel.tireIsInverted then
                            tireDirection = tireDirection * -1;
                        end;
						-------------------------------------------------------------------------------
						if additionalWheel.mudFXtrackIndex ~= nil and ((wetGround and mudFX:getIsTireOnDirt(additionalWheel)) or (additionalWheel.mudFXPuddle ~= nil and getVisibility(additionalWheel.mudFXPuddle.node))) and not mudFX:getIsTireOnMud(additionalWheel,0) then
							--wx, wy, wz = worldToLocal(wheel.node, getWorldTranslation(additionalWheel.wheelTire));
							--wx, wy, wz = localToWorld(wheel.node, wheel.netInfo.x+wheel.xOffset,wheel.netInfo.y-wheel.radius,wheel.netInfo.z-wheel.radius);
							g_currentMission.tireTrackSystem:addTrackPoint(additionalWheel.mudFXtrackIndex, wx, wy, wz, ux, uy, uz, r, g, b, wheel.dirtAmount, a, tireDirection);
						else						
						-------------------------------------------------------------------------------
							g_currentMission.tireTrackSystem:addTrackPoint(additionalWheel.tireTrackIndex, wx, wy, wz, ux, uy, uz, r, g, b, wheel.dirtAmount, a, tireDirection);
							if additionalWheel.mudFXtrackIndex ~= nil then g_currentMission.tireTrackSystem:cutTrack(additionalWheel.mudFXtrackIndex); end;
						end;
                    end;
                end;
            end;
        else
            g_currentMission.tireTrackSystem:cutTrack(wheel.tireTrackIndex);
			-------------------------------------------------------------------------------
			if self.mudFXMudTrackIndex ~= nil then g_currentMission.tireTrackSystem:cutTrack(self.mudFXMudTrackIndex); end;
			if wheel.mudFXtrackIndex ~= nil then g_currentMission.tireTrackSystem:cutTrack(wheel.mudFXtrackIndex); end;
			-------------------------------------------------------------------------------
            if wheel.additionalWheels ~= nil then
                for _, additionalWheel in pairs(wheel.additionalWheels) do
                    if additionalWheel.tireTrackIndex ~= nil then
                        g_currentMission.tireTrackSystem:cutTrack(additionalWheel.tireTrackIndex);
						-------------------------------------------------------------------------------
						if additionalWheel.mudFXtrackIndex ~= nil then g_currentMission.tireTrackSystem:cutTrack(additionalWheel.mudFXtrackIndex); end;
						-------------------------------------------------------------------------------
                    end;
                end;
            end;
        end;
    end;
end;

Wheels.updateWheelFriction = function (self, wheel, dt)
    if self.isServer then
        local isOnField = wheel.densityType ~= 0
        local depth = wheel.lastColor[4]
        local groundType = WheelsUtil.getGroundType(isOnField, wheel.contact ~= Wheels.WHEEL_GROUND_CONTACT, depth)
        local coeff = WheelsUtil.getTireFriction(wheel.tireType, groundType, g_currentMission.environment.weather:getGroundWetness())
		if (wheel.mudFXisActive ~= nil and wheel.mudFXisActive) and self:getBrakeForce() <=0 then
			local i = 0.2;
			if wheel.additionalWheels ~= nil then i = i + (i*#wheel.additionalWheels); end;
			if i > 0.9 then i = 0.9; end;
			coeff = coeff*i;
		end;
        if self:getLastSpeed() > 0.2 then
            if coeff ~= wheel.tireGroundFrictionCoeff then
                wheel.tireGroundFrictionCoeff = coeff;
                self:setWheelTireFrictionDirty(wheel);
            end;
        end;
    end;
end;

--[[Washable.getDirtMultiplier = function (self)
    local spec = self.spec_washable
    local multiplier = 1
    if self:getLastSpeed() < 1 then
        multiplier = 0
    end
    if self:getIsOnField() then
        multiplier = multiplier * spec.fieldMultiplier
        local wetness = g_currentMission.environment.weather:getGroundWetness()
        if wetness > 0 then
            multiplier = multiplier * (1 + wetness)
        end
		if mudFX:getIsGroundWet() then multiplier = multiplier * 10; end;
    end	
    return multiplier
end]]

Wheels.onUpdateTick = function (self, dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    local spec = self.spec_wheels;
    for _, wheel in pairs(spec.wheels) do
        if wheel.rotSpeedLimit ~= nil then
            local dir = -1;
            if self:getLastSpeed() <= wheel.rotSpeedLimit then
                dir = 1;
            end;
            wheel.currentRotSpeedAlpha = MathUtil.clamp(wheel.currentRotSpeedAlpha + dir*(dt/1000), 0, 1);
            wheel.rotSpeed    = wheel.rotSpeedDefault * wheel.currentRotSpeedAlpha;
            wheel.rotSpeedNeg = wheel.rotSpeedNegDefault * wheel.currentRotSpeedAlpha;
        end;
	end;
	if self.isClient then
        local speed = self:getLastSpeed();
        local groundWetness = g_currentMission.environment.weather:getGroundWetness();
        local groundIsWet = groundWetness > 0.2;
        for _,wheel in pairs(spec.wheels) do
			if wheel.driveGroundParticleSystems ~= nil then
				local states = wheel.driveGroundParticleStates;
				local enableSoilPS = false;
				if wheel.lastTerrainValue > 0 and wheel.lastTerrainValue < 5 then
					enableSoilPS = (speed > 1);
				end;
				local sizeScale = 2 * wheel.width * wheel.radiusOriginal;
				states.driving_dry = enableSoilPS;
				states.driving_wet = enableSoilPS and groundIsWet;
				states.driving_dust = not groundIsWet;
				for psName, state in pairs(states) do
					local typedPs = wheel.driveGroundParticleSystems[psName];
					for _, ps in ipairs(typedPs) do
						if wheel.mudFXisActive ~= nil and wheel.mudFXisActive then state = false; end;
						if state then
							if self.movingDirection < 0 then
								setRotation(ps.emitterShape, 0, math.pi+wheel.steeringAngle, 0);
							else
								setRotation(ps.emitterShape, 0, wheel.steeringAngle, 0);
							end;
							local scale;
							if psName ~= "driving_dust" then
								local wheelSpeed = MathUtil.rpmToMps(wheel.netInfo.xDriveSpeed / (2*math.pi) * 60, wheel.radius);
								local wheelSlip = math.pow(wheelSpeed/self.lastSpeedReal, 2.5);
								scale = self:getDriveGroundParticleSystemsScale(ps, wheelSpeed) * wheelSlip;
							else
								scale = self:getDriveGroundParticleSystemsScale(ps, self.lastSpeedReal);
							end;
							if ps.isTintable then
								if ps.lastColor == nil then
									ps.lastColor = {ps.wheel.lastColor[1],ps.wheel.lastColor[2],ps.wheel.lastColor[3]};
									ps.targetColor = {ps.wheel.lastColor[1],ps.wheel.lastColor[2],ps.wheel.lastColor[3]};
									ps.currentColor = {ps.wheel.lastColor[1],ps.wheel.lastColor[2],ps.wheel.lastColor[3]};
									ps.alpha = 1;
								end;
								if ps.alpha ~= 1 then
									ps.alpha = math.min(ps.alpha + dt/1000, 1);
									ps.currentColor = {MathUtil.vector3ArrayLerp(ps.lastColor, ps.targetColor, ps.alpha)};
									if ps.alpha == 1 then
										ps.lastColor[1] = ps.currentColor[1];
										ps.lastColor[2] = ps.currentColor[2];
										ps.lastColor[3] = ps.currentColor[3];
									end;
								end;
								if ps.alpha == 1 and ps.wheel.lastColor[1] ~= ps.targetColor[1] and ps.wheel.lastColor[2] ~= ps.targetColor[2] and ps.wheel.lastColor[3] ~= ps.targetColor[3] then
									ps.alpha = 0;
									ps.targetColor[1] = ps.wheel.lastColor[1];
									ps.targetColor[2] = ps.wheel.lastColor[2];
									ps.targetColor[3] = ps.wheel.lastColor[3];
								end;
							end;
							if scale > 0 then
								ParticleUtil.setEmittingState(ps, true);
								if ps.isTintable then
									setShaderParameter(ps.shape, "psColor", ps.currentColor[1], ps.currentColor[2], ps.currentColor[3], 1, false);
								end;
							else
								ParticleUtil.setEmittingState(ps, false);
							end;
							local maxSpeed = (50 / 3.6);
							local circum = wheel.radiusOriginal;
							local maxWheelRpm = maxSpeed / circum;
							local wheelRotFactor = Utils.getNoNil(wheel.netInfo.xDriveSpeed, 0) / maxWheelRpm;
							local emitScale = scale * wheelRotFactor * sizeScale;
							ParticleUtil.setEmitCountScale(ps, MathUtil.clamp(emitScale, ps.minScale, ps.maxScale));
							local speedFactor = 1.0;
							ParticleUtil.setParticleSystemSpeed(ps, ps.particleSpeed * speedFactor);
							ParticleUtil.setParticleSystemSpeedRandom(ps, ps.particleRandomSpeed * speedFactor);
						else
							ParticleUtil.setEmittingState(ps, false);
						end;
					end;
					states[psName] = false;
				end;
			end;			
        end;
    end;
end;